//
// Copyright Contributors to the MaterialX Project
// SPDX-License-Identifier: Apache-2.0
//

#ifndef MATERIALX_SURFACENODEMSL_H
#define MATERIALX_SURFACENODEMSL_H

#include <MaterialXGenHw/Nodes/HwSurfaceNode.h>

MATERIALX_NAMESPACE_BEGIN

// Backward compatibility header for OpenUSD and other external
// projects referencing files that were removed in MaterialX 1.39.5
using SurfaceNodeMsl = HwSurfaceNode;

MATERIALX_NAMESPACE_END

#endif
