//
// Copyright Contributors to the MaterialX Project
// SPDX-License-Identifier: Apache-2.0
//

#ifndef MATERIALX_GLUTIL_H
#define MATERIALX_GLUTIL_H

/// @file
/// OpenGL utilities

#include <MaterialXRenderGlsl/Export.h>

#include <MaterialXCore/Library.h>

MATERIALX_NAMESPACE_BEGIN

MX_RENDERGLSL_API void checkGlErrors(const string& context);

MATERIALX_NAMESPACE_END

#endif
